from jinja2 import Environment, FileSystemLoader
import click
from .config import BeamConfig, PermanentSettings
from .run import run
from .settings import *

beam_version = "(no version because beam is not packaged)"

try:
    from tca_beam import __version__
    beam_version = __version__
except ImportError:
    pass

# TODO:
# [x] try trogon/textualize on tca-beam -- see results in my main TW -- might have to make explicit tca-beam 'make' command
# -[x] upload to main pyPi
# -[ ] make bbtests
# -[x] option for preview-all
# -[x] impl two files
# -[x] flags for 'features in sub_dirs'
# -[x] make script use abs path to the templates, no chdir!
# -[x] --dry-run

# For docs:
#  tca-beam won't complain if the output_dir already exists, regardless of force_overwrite flag (which only applies to files).
#

# When you're testing and playing with beam, it might be tempting to set up a host xcode project with a _link_ to a folder
# where you are generating the files with beam.
# However, Xcode however won't add these files to any target (it's how linked folders work), so you're probably better off
# just adding a folder group to your project (in the usual way) after you've generated the files with beam.


@click.command(no_args_is_help=True)
@click.option('--two-files', is_flag=True, help="Put view and reducer into separate files")
@click.option('--sub-dirs', is_flag=True, help="Put each feature in a sub-directory")
@click.option('--preview-all', is_flag=True, help="Generate a single View that previews all feature Views")
@click.option('--output-dir', default='.', help="Output directory (defaults to current dir)")
@click.option('--force-overwrite', is_flag=True, help="Force overwriting any existing files")
@click.option('--dry-run', is_flag=True, help="Don't generate files, just preview any actions")
@click.option('--customise-settings', is_flag=True, help="Generate a user-editable file to tweak file naming settings.")
@click.option('--version', is_flag=True, help="Print version and exit")
@click.argument('feature_names', nargs=-1)
def start(two_files, sub_dirs, preview_all, output_dir, force_overwrite, dry_run, customise_settings, version, feature_names):

    if customise_settings:
        personalize_permanent_settings()
        sys.exit(0)

    permanent_settings = load_permanent_settings()
    dbg(f"Settings: {permanent_settings}")

    if version:
        p(beam_version)
        sys.exit(0)

    script_dir = os.path.abspath(os.path.dirname(__file__))

    templates_path = make_abs_path('templates')
    file_loader = FileSystemLoader(templates_path)
    jinja_env = Environment(loader=file_loader)

    config = BeamConfig(PermanentSettings(permanent_settings), script_dir, output_dir, jinja_env, two_files, sub_dirs, preview_all,
                        output_dir, force_overwrite, dry_run, feature_names)

    if len(feature_names) < 1:
        p()
        p("Please give one or more feature name arguments (after any option flags).")
        p()

        echo = click.echo
        echo(start.get_help(click.get_current_context()))
        sys.exit(1)

    p("")
    p("tca-beam is preparing two-by-fours...")
    p("")

    # if dry_run:
    #     p("In DRY-RUN mode. No files or folders will be created.")
    #     p("")

    # the output_dir is relative to the user's current dir, NOT the script!

    run(config)


if __name__ == '__main__':
    start()



import sys
from dataclasses import dataclass
from jinja2 import Environment
from .helpers import *


def print_settings_advice_and_exit():
    p("""            
    Please check the file ~/.beam-settings.toml for issues or corruption,
    and consider deleting it and running 'tca-beam --customise-settings'
    to generate the file again with the default values.
    
    """)

    sys.exit(1)


class PermanentSettings:

    def __init__(self, configParser):
        self.configParser = configParser
        self.two_files_view_part_filename = self.sanitized_setting('two_files_view_part_filename')
        self.two_files_reducer_part_filename = self.sanitized_setting('two_files_reducer_part_filename')
        self.one_file_filename = self.sanitized_setting('one_file_filename')
        self.preview_all_filename = self.sanitized_setting('preview_all_filename')
        self.feature_subdirs_dir_name = self.sanitized_setting('feature_subdirs_dir_name')


    def sanitized_setting(self, setting_name) -> str:
        section = "file_naming"

        setting_string = ""

        try:
            setting_string = self.configParser.get(section, setting_name)
        except Exception as e:
            error(f"Error reading setting '{setting_name}' from the settings file.\n")
            print_settings_advice_and_exit()

        if not setting_string or len(setting_string) <= 0:
            error("Couldn't find a key {setting_name} in the ")
            print_settings_advice_and_exit()

        # check for bungled string replacement in settings
        if ("{" in setting_string or "}" in setting_string) and "{{featureName}}" not in setting_string:
            p(f"""

ERROR: Found a user-customised settings string for key '{setting_name}' that seems to contain a corrupt tag:

    {setting_string}

The only tag you can use is '{{{{featureName}}}}'.
""")
            print_settings_advice_and_exit()

        dbg(f"Got sanitized settings: {setting_string}")

        return sanitize_filename(setting_string)


@dataclass
class BeamConfig:
    permanent_settings: PermanentSettings
    script_dir: str
    target_dir: str
    jinja_env: Environment
    two_files: bool
    sub_dirs: bool
    preview_all: bool
    output_dir: str
    force_overwrite: bool
    dry_run: bool
    feature_names: [str]



import sys
import shutil
from configparser import ConfigParser
from pathlib import Path
from typing import Optional
from .helpers import *

settings_filename = '.beam-settings.toml'
settings_user_copy_filename = settings_filename
settings_built_in_path = f"templates/{settings_filename}"


def personalize_permanent_settings(force=False):
    """
    Copies the default settings file to the user's home directory.
    """
    # location of the default settings file
    default_settings_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), settings_built_in_path)

    # destination for the copied settings file in the user's home directory
    home = str(Path.home())
    destination_settings_path = os.path.join(home, settings_filename)

    if not force and os.path.exists(destination_settings_path):
        p()
        p(f"A settings file already exists at {destination_settings_path}.")
        p("If you want to generate the settings file again from defaults, please delete that existing file first.")
        p()
        sys.exit(1)

    # Copy the settings file
    shutil.copyfile(default_settings_path, destination_settings_path)

    p(f"I've copied default settings to {destination_settings_path}.")
    p(f"Please edit this file with your favourite text editor.")
    p()
    sys.exit(0)


def load_permanent_settings() -> Optional[ConfigParser]:
    """
    Loads the user's custom settings if they exist, otherwise loads the default settings.
    Returns a ConfigParser object, or None if no settings file is found.
    """
    # location of the settings files
    home = str(Path.home())
    user_settings_path = os.path.join(home, settings_user_copy_filename)
    default_settings_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), settings_built_in_path)

    dbg(f"got user_settings_path = {user_settings_path}")
    dbg(f"got default_settings_path = {default_settings_path}")

    configParser = ConfigParser()

    # If the user's custom settings exist, load them
    if os.path.exists(user_settings_path):
        configParser.read(user_settings_path)
    # ... or if the default settings exist, load them
    elif os.path.exists(default_settings_path):
        configParser.read(default_settings_path)
    else:
        return None

    return configParser



from .template_rendering import *


def process_template(config, feature_name):
    dbg(f"start process_template, config = {config}, sub_dirs = {config.sub_dirs}")

    substitutions = {
        'viewName': f"{feature_name}View",
        'featureName': f"{feature_name}ViewFeature"
    }

    # Load the template
    view_template = config.jinja_env.get_template('View.swift')
    view_feature_template = config.jinja_env.get_template('ViewFeature.swift')

    view_content = view_template.render(substitutions)
    view_feature_content = view_feature_template.render(substitutions)

    substitutions['reducerContent'] = view_feature_content
    substitutions['viewContent'] = view_content

    template_renders = []

    target_dir = f"{feature_name}Feature" if config.sub_dirs else "."

    dbg(f"just made target_dir: {config.target_dir}, and sub_dirs == {config.sub_dirs}")
    if config.two_files:
        # could do replacement inside TemplateRender? Or make method helper in this file, more likely
        reducer_filename = config.permanent_settings.two_files_reducer_part_filename.replace("{{featureName}}", feature_name)
        dbg(f"Made RF: {reducer_filename}")
        template_renders.append(TemplateRender(reducer_filename,
                                               config.jinja_env.get_template('TwoFile_ReducerPart.swift'),
                                               target_dir))

        view_filename = config.permanent_settings.two_files_view_part_filename.replace("{{featureName}}", feature_name)
        dbg(f"Made VF: {view_filename}")
        template_renders.append(TemplateRender(view_filename,
                                               config.jinja_env.get_template('TwoFile_ViewPart.swift'),
                                               target_dir))
    else:
        one_file_filename = config.permanent_settings.one_file_filename.replace("{{featureName}}", feature_name)
        dbg(f"Made OFF: {one_file_filename}")
        # for reducer + view in one file, we append just 'View' to feature name
        template_renders.append(TemplateRender(one_file_filename,
                                               config.jinja_env.get_template('OneFile.swift'),
                                               target_dir))

    step_name = f"Feature '{feature_name}':"
    render_templates(config, template_renders, substitutions, step_name)


def generate_all_preview(config):

    # a single View that has a preview for all the Views
    all_previews_substitutions = []

    for feature_name in config.feature_names:
        all_previews_substitutions.append({
            'viewName': f"{feature_name}View",
            'featureName': f"{feature_name}ViewFeature"
        })

    substitutions_all_previews = { 'allFeatures': all_previews_substitutions}

    preview_all_filename = config.permanent_settings.preview_all_filename.replace("{{featureName}}", feature_name)

    template_render = TemplateRender(preview_all_filename,
                                     config.jinja_env.get_template('AllPreviews.swift'))

    render_templates(config, [template_render], substitutions_all_previews, 'Preview for all features:')


def run(config):
    for feature_name in config.feature_names:
        process_template(config, feature_name)

    if config.preview_all:
        generate_all_preview(config)

    p()
    p("Done")
    p()



import sys
from collections import namedtuple
from pathlib import Path
from .helpers import *

TemplateRender = namedtuple('TemplateRender', ['render_file', 'template', 'target_dir'],
                            defaults=['', '', '.'])


def render_templates(config, templateRenders, substitutions, step_name):

    console_prefix = "- (DRY RUN:) " if config.dry_run else "- "

    dbg(f"-------- render_templates: templateRenders = {templateRenders}")
    dbg(f"-------- render_templates: substitutions = {substitutions}")

    p(f"{console_prefix}{step_name}")

    for templateRender in templateRenders:

        stub_contents = templateRender.template.render(substitutions)

        # Open the file for writing
        filename = templateRender.render_file

        structure_directory = Path(templateRender.target_dir)

        abs_directory = config.output_dir / structure_directory
        filepath = abs_directory / filename

        p(f"{console_prefix}   Creating file {filepath}")

        if not config.dry_run:
            if os.path.isdir(filepath):
                error(f'A directory named "{filepath}" already exists, refusing to overwrite.')
                p()
                sys.exit(1)

            if not config.force_overwrite and os.path.isfile(filepath):
                error(f'A file named "{filepath}" already exists, refusing to overwrite. Use --force-overwrite to suppress this error.')
                p()
                sys.exit(1)

            abs_directory.mkdir(parents=True, exist_ok=True)

            with open(filepath, 'w') as f:
                f.write(stub_contents + '\n')



import os
import re


def dbg(string):
    pass
    # print(string)


def p(string=""):
    print(string)


def error(string):
    print(f"\n ERROR: {string}")


def make_abs_path(rel_path):
    script_dir = os.path.dirname(__file__)
    return os.path.normpath(os.path.join(script_dir, rel_path))


def sanitize_filename(filename):
    return re.sub(r'(?u)[^-\w.{}]', '', filename.strip())
